#pragma once
#include <cstddef>
#include <exception>

#include "smart_ptr/ptr_detail.hpp"

namespace framework
{
	//
	// framework::smart_ptr<T> - QƃJEgX}[g|C^ v3.0.0.0
	//
	template<class T> class smart_ptr
	{
	private:
		detail::ptr_inner* _inner;
	public:
		// smart_ptr<T> ܂B
		smart_ptr() : _inner(nullptr) {}
		smart_ptr(const smart_ptr<T>& c) : _inner(nullptr) { this->Copy(c); }
		explicit smart_ptr(detail::null_ptr& n) : _inner(nullptr) {}
		template<class U> explicit smart_ptr(U* object) : _inner(nullptr) {this->Reset(object, detail::func_ptr_delete<U>); }
		template<class U, class D> explicit smart_ptr(U* object, D deleter) : _inner(nullptr) {this->Reset(object, deleter); }
		// smart_ptr<T> j܂B
		~smart_ptr() { this->Free(); }
	public:
		// IuWFNg̃|C^擾܂B
		T* Get() const { return this->Get<T>();  }
		template<class U> U* Get() const { return static_cast<U*>(_inner ? _inner->object : nullptr); }
		// ݂̎QƃJEg擾܂B
		uintptr_t Count() const { return _inner ? _inner->count : nullptr; }
	public:
		// ǗĂIuWFNg܂B
		inline void Free();
		// ǗĂIuWFNgL܂B
		inline void Copy(const smart_ptr<T>& from);
		// IuWFNgւ܂B
		inline void Swap(smart_ptr<T>& swp);
		// CX^XɐVIuWFNgݒ肵܂B
		template<class U> void Reset(U* object) { this->Reset(object, detail::func_ptr_delete<U>); }
		template<class U, class D> inline void Reset(U* object, D deleter);
		// IuWFNgLXg܂B
		template<class U> inline smart_ptr<U> Cast() const { return ptr_cast<U>(*this); }
	public:
		// Zq = 񋟂܂B
		smart_ptr<T>& operator =(const smart_ptr<T>& rhs)		{ this->Copy(rhs); return *this; }
		smart_ptr<T>& operator =(const detail::null_ptr& rhs)	{ this->Free(); return *this; }
		// ̑̉Zq񋟂܂B
		T* operator ->() const	{ return this->Get(); }
		T& operator *()	const	{ return *this->Get(); }
		operator bool()	const	{ return this->Get() != nullptr; }
	public:
		// rZq == 񋟂܂B
		bool operator ==(const smart_ptr<T>& rhs) const		{ return this->Get() == rhs.Get(); }
		bool operator ==(const T* rhs) const				{ return this->Get() == rhs; }
		bool operator ==(const detail::null_ptr& rhs) const	{ return this->Get() == nullptr; }
		// rZq != 񋟂܂B
		bool operator !=(const smart_ptr<T>& rhs) const		{ return this->Get() != rhs.Get(); }
		bool operator !=(const T* rhs) const				{ return this->Get() != rhs; }
		bool operator !=(const detail::null_ptr& rhs) const	{ return this->Get() != nullptr; }
		// rZq <= 񋟂܂B
		bool operator <=(const smart_ptr<T>& rhs) const		{ return this->Get() <= rhs.Get(); }
		bool operator <=(const T* rhs) const				{ return this->Get() <= rhs; }
		bool operator <=(const detail::null_ptr& rhs) const	{ return this->Get() <= nullptr; }
		// rZq >= 񋟂܂B
		bool operator >=(const smart_ptr<T>& rhs) const		{ return this->Get() >= rhs.Get(); }
		bool operator >=(const T* rhs) const				{ return this->Get() >= rhs; }
		bool operator >=(const detail::null_ptr& rhs) const	{ return this->Get() >= nullptr; }
		// rZq < 񋟂܂B
		bool operator <(const smart_ptr<T>& rhs) const		{ return this->Get() < rhs.Get(); }
		bool operator <(const T* rhs) const					{ return this->Get() < rhs; }
		bool operator <(const detail::null_ptr& rhs) const	{ return this->Get() < nullptr; }
		// rZq > 񋟂܂B
		bool operator >(const smart_ptr<T>& rhs) const		{ return this->Get() > rhs.Get(); }
		bool operator >(const T* rhs) const					{ return this->Get() > rhs; }
		bool operator >(const detail::null_ptr& rhs) const	{ return this->Get() > nullptr; }
	public:
		// X}[g|C^LXg܂B
		template<class U, class V> friend inline smart_ptr<U> ptr_cast(const smart_ptr<V>& obj);
	};

	// X}[g|C^ɂ null|C^
	static detail::null_ptr null;
}

namespace framework
{
	// VX}[g|C^쐬܂B
	template<class T> smart_ptr<T> new_ptr(T* object)
	{
		return smart_ptr<T>(object);
	}

	// 폜ȂÓI|C^쐬܂B
	template<class T> smart_ptr<T> static_ptr(T* object)
	{
		return smart_ptr<T>(object, detail::func_ptr_static);
	}
}

// smart_ptr-impl.hpp ̃CN[h
#include "smart_ptr/smart_ptr-impl.hpp"
