#pragma once
#include <cstdio>
#include <windows.h>
#include <cstdlib>
#include <mmsystem.h>
#include <boost/shared_ptr.hpp>
#include "smart_ptr.hpp"
#pragma comment(lib, "winmm.lib")

const int sampling_count = 10;
const int loop_count = 100000;

namespace boost_sc
{
	

	int score_instance()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++) boost::shared_ptr<byte> p(new byte(0));
		end = timeGetTime();
		return end -begin;
	}

	int score_ret()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++)
		{
			boost::shared_ptr<byte> p(new byte(0));
			boost::shared_ptr<byte> ret = p;
		}
		end = timeGetTime();
		return end -begin;
	}
	
	int score_reset()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++)
		{
			boost::shared_ptr<byte> p(new byte(0));
			p.reset(new byte(0));
		}
		end = timeGetTime();
		return end -begin;
	}

	int score_copy()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++)
		{
			boost::shared_ptr<byte> p(new byte(0));
			boost::shared_ptr<byte> p2(new byte(0));
			p = p2;
		}
		end = timeGetTime();
		return end -begin;
	}

	int score_free()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++)
		{
			boost::shared_ptr<byte> p(new byte(0));
			p.reset();
		}
		end = timeGetTime();
		return end -begin;
	}
}

namespace fw_sc
{
	int score_instance()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++) framework::smart_ptr<byte> p(new byte(0));
		end = timeGetTime();
		return end -begin;
	}

	int score_ret()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++)
		{
			framework::smart_ptr<byte> p(new byte(0));
			framework::smart_ptr<byte> ret = p;
		}
		end = timeGetTime();
		return end -begin;
	}
	
	int score_reset()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++)
		{
			framework::smart_ptr<byte> p(new byte(0));
			p.Reset(new byte(0));
		}
		end = timeGetTime();
		return end -begin;
	}

	int score_copy()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++)
		{
			framework::smart_ptr<byte> p(new byte(0));
			framework::smart_ptr<byte> p2(new byte(0));
			p = p2;
		}
		end = timeGetTime();
		return end -begin;
	}

	int score_free()
	{
		int begin, end;
		begin = timeGetTime();
		for (int i = 0; i < loop_count; i++)
		{
			framework::smart_ptr<byte> p(new byte(0));
			p.Free();
		}
		end = timeGetTime();
		return end -begin;
	}
}

struct calc_result
{
	int instance;
	int ret;
	int reset;
	int copy;
	int free;

	calc_result() : instance(0), ret(0), reset(0), copy(0), free(0) {}
};

void calc_boost(calc_result& res)
{
	for (int i = 0; i < sampling_count; i++)
	{
		res.instance += boost_sc::score_instance();
		res.ret += boost_sc::score_ret();
		res.reset += boost_sc::score_reset();
		res.copy += boost_sc::score_copy();
		res.free += boost_sc::score_free();
	}
	res.instance = res.instance / sampling_count;
	res.ret = res.ret / sampling_count;
	res.reset = res.reset / sampling_count;
	res.copy = res.copy / sampling_count;
	res.free = res.free / sampling_count;
}

void calc_smart(calc_result& res)
{
	for (int i = 0; i < sampling_count; i++)
	{
		res.instance += fw_sc::score_instance();
		res.ret += fw_sc::score_ret();
		res.reset += fw_sc::score_reset();
		res.copy += fw_sc::score_copy();
		res.free += fw_sc::score_free();
	}
	res.instance = res.instance / sampling_count;
	res.ret = res.ret / sampling_count;
	res.reset = res.reset / sampling_count;
	res.copy = res.copy / sampling_count;
	res.free = res.free / sampling_count;
}
